# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
import os
from time import time

IMPALA_AUDIT_WAL_FILE_NAME = "impalad_audit_wal"
IMPALA_AUDIT_WAL_TMP_FILE_NAME = "impalad_audit_wal.tmp"
IMPALA_AUDIT_WAL_CHECKPOINT_SIZE_KB = 512
IMPALA_CHECKPOINT_INTERVAL_SECONDS = 5 * 60
SEEK_END = 2
SEPARATOR = ","

class Wal(object):
  """
  A class to manage Write After Log that is used to save the last audit record
  that was successfully sent to audit server. The file is used to resume
  operation after agent is restarted. Each line the file has format:
    _offset,_file_name
  The _file_name is used instead of full path as it helps in moving audit log
  directory along with all the audit logs in it. 
  """

  def __init__(self, log_dir):
    self._offset = None
    self._file_name = None
    self._walfd = None
    self._last_checkpoint_time = time()
    
    self._log_dir = log_dir
    self._wal_file = os.path.join(self._log_dir, IMPALA_AUDIT_WAL_FILE_NAME)
    if os.path.exists(self._wal_file):
      self._walfd = open(self._wal_file, 'a+')
      self._file_name, self._offset = self._init_file_info()
      # checkpoint in case we are recovering from a crash
      if self._file_name is None:
        self._walfd.truncate(0)
      else:
        self._checkpoint(self._file_name, self._offset)
  
  def close(self):
    """
    Close the open fd to the Wal file if it is not None.
    """    
    if self._walfd is not None:
      self._walfd.close()
    
  def _init_file_info(self):
    """
    Read the WAL file for a valid wal line and initialize the audit log file and
    its offset.
    A valid line in WAL file is terminated by "\n". If the line isn't newline
    terminated we assume that something crashed resulting in partial line
    written to the file. 
    """
    file_name, offset = None, None
    
    lines = self._walfd.readlines()
    while lines:
      last_line = lines.pop()
      if self._is_valid_wal_line(last_line):
        offset, file_name = last_line.strip().split(SEPARATOR)
        offset = int(offset)
        break
      
    return file_name, offset
  
  def _is_valid_wal_line(self, line):
    """
    Checks validity of a line read from wal file and if the format isn't correct
    returns False. If the line is correctly formatted, returns True.
    """
    return line and line[-1] == "\n" 
    
  def get_file_info(self):
    """
    Return the information about the current audit log file and _offset in it
    from where we should start collecting audit events.
    """
    file_name = None
    if self._file_name is not None:
      file_name = self._file_name
    
    offset = None
    if self._offset is not None:
      offset = int(self._offset)
    return file_name, offset
  
  def append(self, filename, offset):
    """
    Add information about audit event that was just sent to the audit server.
    Offset is the next audit line that needs to be sent to server. 
    """
    if self._walfd is None:
      self._walfd = open(self._wal_file, 'a+')
    
    if not self._checkpoint_if_needed(filename, offset):
      print >> self._walfd, "%d%s%s" % (offset, SEPARATOR, filename)
      self._walfd.flush()
      os.fsync(self._walfd)

    self._file_name = filename
    self._offset = offset
    
  def _checkpoint_if_needed(self, filename, offset):
    """
    Method to checkpoint current state in the audit log periodically. The
    checkpointing is only performed if the WAL file is more than
    IMPALA_AUDIT_WAL_CHECKPOINT_SIZE_KB in size.
    Returns a boolean indicating if checkpointing condition was met and if
    the checkpointing was performed.
    """
    checkpoint_done = False
    current_time = time()
    if current_time - self._last_checkpoint_time > IMPALA_CHECKPOINT_INTERVAL_SECONDS:
      self._last_checkpoint_time = current_time
      size_in_bytes = os.path.getsize(self._wal_file)
      if size_in_bytes > IMPALA_AUDIT_WAL_CHECKPOINT_SIZE_KB * 1024:
        self._checkpoint(filename, offset)
        checkpoint_done = True
        
    return checkpoint_done
  
  def _checkpoint(self, filename, offset):
    """
    Method to checkpoint current state in the audit log. The method creates
    a new file and writes the current state in it. It then renames this file
    to the wal file (relies on posix guarantee of atomic rename)
    """
    tmp_wal_file = os.path.join(self._log_dir, IMPALA_AUDIT_WAL_TMP_FILE_NAME)
    
    # write current file offset to the file and then rename
    tmp_wal = open(tmp_wal_file, 'w')
    print >> tmp_wal, "%d%s%s" % (offset, SEPARATOR, filename)
    tmp_wal.close()
    
    # rename the file to _wal_file
    self._walfd.close()
    os.rename(tmp_wal_file, self._wal_file)
    self._walfd = open(self._wal_file, 'a+')
